/**
  *  \file client/dialogs/fleetopts.cc
  *  \brief Fleet Cost Options dialog
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include "client/dialogs/fleetopts.h"
#include "client/widgets/helpwidget.h"
#include "client/widgets/optiongrid.h"
#include "cpluslib/misc.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "ui/widgets/button.h"
#include "util/nls.h"
#include "util/strtab.h"
#include "ui/widgets/quit.h"

namespace {
    class WFleetCostOptionsTile : public WOptionGrid {
     public:
        WFleetCostOptionsTile(WFleetCostOptions& opts);

     private:
        void init();
        void render();
        void onClick(int id);

        WFleetCostOptions& opts;
    };
}

static const char*const left_side[] = {
    N_("Fighter cost"),
    N_("Include torpedoes"),
    N_("Include engines"),
    N_("Include planet defense"),
    N_("Include starbase"),
    N_("Include starbase tech levels"),
    N_("Include ship tech levels"),
    N_("Display by"),
};

static const char*const right_side[] = {
    N_("no"),                    // 0 = false
    N_("yes"),                   // 1 = true
    N_("player"),                // 2 = 2+false
    N_("team"),                  // 3 = 2+true
    N_("not included"),          // 4 = 4+ff_Free
    N_("built by \"lfm\""),      // 5 = 4+ff_LFM
    N_("built on starbase"),     // 6 = 4+ff_Build
    N_("no"),                    // 7 = 7+st_No
    N_("once per player"),       // 7 = 7+st_Player
    N_("once per ship"),         // 8 = 7+st_Ship
};

static const char keys[] = { 'F', 'T', 'E', 'P', 'B', 'L', 'S', 'Y' };

WFleetCostOptionsTile::WFleetCostOptionsTile(WFleetCostOptions& opts)
    : WOptionGrid(getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(left_side, countof(left_side))),
                  getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(right_side, countof(right_side)))),
      opts(opts)
{
    init();
}

void
WFleetCostOptionsTile::init()
{
    CTASSERT(countof(keys) == countof(left_side));
    for (size_t i = 0; i < countof(left_side); ++i) {
        addItem(keys[i], keys[i], _(left_side[i]));
    }
    sig_click.add(this, &WFleetCostOptionsTile::onClick);
    render();
}

void
WFleetCostOptionsTile::onClick(int id)
{
    switch (id) {
     case 'F':
        opts.fighters = WFleetCostOptions::Fighters(opts.fighters + 1);
        if (opts.fighters >= WFleetCostOptions::ff_MAX)
            opts.fighters = WFleetCostOptions::ff_Free;
        render();
        break;
     case 'T':
        opts.torps = !opts.torps;
        render();
        break;
     case 'E':
        opts.engine = !opts.engine;
        render();
        break;
     case 'P':
        opts.planet = !opts.planet;
        render();
        break;
     case 'B':
        opts.base = !opts.base;
        render();
        break;
     case 'L':
        opts.tech = !opts.tech;
        render();
        break;
     case 'S':
        opts.shiptech = WFleetCostOptions::ShipTech(opts.shiptech + 1);
        if (opts.shiptech >= WFleetCostOptions::st_MAX) {
            opts.shiptech = WFleetCostOptions::st_No;
        }
        render();
        break;
     case 'Y':
        opts.team = !opts.team;
        render();
        break;
    }
}

void
WFleetCostOptionsTile::render()
{
    find('F').setValue(_(right_side[4 + opts.fighters]));
    find('T').setValue(_(right_side[opts.torps]));
    find('E').setValue(_(right_side[opts.engine]));
    find('P').setValue(_(right_side[opts.planet]));
    find('B').setValue(_(right_side[opts.base]));
    find('L').setValue(_(right_side[opts.tech]));
    find('S').setValue(_(right_side[7 + opts.shiptech]));
    find('Y').setValue(_(right_side[2 + opts.team]));
}

/** Edit fleet cost options.
    \param opts Option set to edit
    \retval true options were modified, user confirmed the dialog
    \retval false user canceled dialog, options not modified */
bool
editFleetCostOptions(WFleetCostOptions& opts)
{
    WFleetCostOptions new_opts(opts);

    ObjectHolder h;
    UIWindow& win = h.add(new UIWindow(_("Fleet Cost Comparison"), UIWindow::BLUE_WINDOW, COLOR_GRAY));
    win.setLayout(UIVBoxLayout::instance5);
    win.add(h.add(new WFleetCostOptionsTile(new_opts)));

    UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
    g.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
    g.add(h.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    g.add(h.add(new UISpacer()));
    g.add(h.add(new UIButton(_("Help"), 'h')));
    win.add(g);
    win.add(h.add(new WHelpWidget("pcc2:fleetcostopts")));
    win.add(h.add(new UIQuit(0)));
    win.pack();
    ui_root->centerWidget(win);
    if (ui_root->runModal(win)) {
        opts = new_opts;
        return true;
    } else {
        return false;
    }
}
