/**
  *  \file client/dialogs/notifydlg.cc
  *  \brief Notification Message Display
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/notifydlg.h"
#include "client/widgets/helpwidget.h"
#include "client/widgets/msgaction.h"
#include "client/widgets/msgdisplay.h"
#include "cpluslib/holder.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/ship.h"
#include "gfx/palette.h"
#include "int/notify.h"
#include "ui/layout.h"
#include "ui/prefix.h"
#include "ui/uiglobal.h"
#include "ui/widgets/button.h"
#include "ui/window.h"
#include "util/misc.h"
#include "util/nls.h"
#include "client/widgets/decaymsg.h"
#include "client/dialogbinding.h"
#include "int/exec.h"
#include "ui/widgets/quit.h"

namespace {
    /** Implementation of WMessageActionPanel for notification messages. */
    class WNotifyActionPanel : public WMessageActionPanel {
        IntNotificationMessageStore& inms;
     public:
        WNotifyActionPanel(IntNotificationMessageStore& inms)
            : WMessageActionPanel(inms),
              inms(inms)
            {
                onBrowse();
            }
        void updateMessageControls();
    };

    /** Notification message window. */
    class WNotifyWindow : public UIWindow, public IntUserInterfaceDialogBinding {
        IntNotificationMessageStore& inms;
        ObjectHolder h;

        void init();
     public:
        WNotifyWindow(IntNotificationMessageStore& inms);
        bool handleEvent(const UIEvent& event, bool second_pass);
    };
}

void
WNotifyActionPanel::updateMessageControls()
{
    // Check data
    IntNotificationMessage* msg = inms.getMessageByIndex(inms.getCurrent());
    IntExecutionContext::ProcessKind pk =
        msg->getAssociatedProcess() != 0 ? msg->getAssociatedProcess()->getProcessKind() : IntExecutionContext::pkDefault;

    // Check for object, syntactically
    int id = 0;
    int kind = 0;
    string_t s = inms.getText(inms.getCurrent());
    if (s.size() > 3 && s[0] == '(' && s[1] == '-') {
        if (s[2] == 'p') {
            kind = 2;
        } else if (s[2] == 's') {
            kind = 1;
        }
        stringValue(s.substr(3, s.find(')', 3)-3), id);
    }

    // Validate
    if (kind == 1 && current_anyship_type.isValidIndex(id)) {
        if (current_ship_type.isValidIndex(id)) {
            if (pk == IntExecutionContext::pkShipTask) {
                setHaveObject(true, WScreenId(WScreenId::ShipTaskScreen, id));
            } else {
                setHaveObject(true, WScreenId(WScreenId::ShipScreen, id));
            }
        } else {
            setHaveObject(true, WScreenId(dynamic_cast<GMapObject&>(current_anyship_type.getObjectByIndex(id)).getPos()));
        }
    } else if (kind == 2 && current_anyplanet_type.isValidIndex(id)) {
        if (current_planet_type.isValidIndex(id)) {
            if (pk == IntExecutionContext::pkPlanetTask) {
                setHaveObject(true, WScreenId(WScreenId::PlanetTaskScreen, id));
            } else if (pk == IntExecutionContext::pkBaseTask) {
                setHaveObject(true, WScreenId(WScreenId::BaseTaskScreen, id));
            } else {
                setHaveObject(true, WScreenId(WScreenId::PlanetScreen, id));
            }
        } else {
            setHaveObject(true, WScreenId(dynamic_cast<GMapObject&>(current_anyplanet_type.getObjectByIndex(id)).getPos()));
        }
    }

    // Check "confirm" status
    if (msg->getAssociatedProcess() != 0 && !msg->isConfirmed()) {
        setHaveConfirm(true);
    }
}

/***************************** WNotifyWindow *****************************/

/** Constructor. Builds the whole window. */
WNotifyWindow::WNotifyWindow(IntNotificationMessageStore& inms)
    : UIWindow(_("Notifications"), BLUE_WINDOW, COLOR_BLACK),
      inms(inms)
{
    init();
}

void
WNotifyWindow::init()
{
    // HBox
    //   VBox
    //     WMessageActionPanel
    //     HBox
    //       UIButton "Close"
    //       UISpacer
    //   WMessageDisplay
    // FIXME: this dialog currently has no clickable "Help" button.
    setLayout(UIHBoxLayout::instance5);

    UIGroup& g1  = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& g11 = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& g2  = h.add(new UIGroup(UIVBoxLayout::instance5));

    g1.add(h.add(new WNotifyActionPanel(inms)));
    g1.add(g11);

    g11.add(h.add(new UIButton(1, _("Close"), SDLK_ESCAPE)).addStop());
    g11.add(h.add(new UIButton(_("Tab - List"), SDLK_TAB)));
    g11.add(h.add(new UISpacer()));

    g2.add(h.add(new WMessageDisplay(0, inms)));

    add(g1);
    add(g2);
    add(h.add(new UIPrefixArg()));
    add(h.add(new WHelpWidget("pcc2:notify")));
    add(h.add(new UIQuit(1)));

    pack();
}

bool
WNotifyWindow::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case 'c':
            {
                IntNotificationMessage* msg = inms.getMessageByIndex(inms.getCurrent());
                if (msg->getAssociatedProcess() != 0 && !msg->isConfirmed()) {
                    msg->setConfirmed(true);
                    inms.sig_sel_changed.raise();
                }
            }
            return true;
        }
    }
    return UIWindow::handleEvent(event, second_pass);
}

/** Display notification messages. This also allows the user to
    confirm them to continue suspended processes. Therefore, it must
    be followed by a call to runRunnableProcesses(). */
void
doNotifyScreen()
{
    IntNotificationMessageStore& inms = IntNotificationMessageStore::instance;
    if (inms.getNumMessages() == 0) {
        showDecayingMessage(_("No Notification Messages"));
    } else {
        WNotifyWindow win(IntNotificationMessageStore::instance);
        ui_root->centerWidget(win);
        win.runModal(win);
        IntNotificationMessageStore::instance.makeConfirmedProcessesRunnable();
    }
}
