/**
  *  \file int/if/ufoif.cc
  *  \brief Interpreter Game Interface: Ufo Properties
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/if/ufoif.h"
#include "cpluslib/format.h"
#include "game/tables.h"
#include "game/ufo.h"
#include "int/interror.h"
#include "int/value.h"
#include "int/if/ifutil.h"
#include "int/data.h"
#include "game/global.h"
#include "cpluslib/misc.h"
#include "int/compound.h"
#include "int/args.h"
#include "int/if/objif.h"
#include "int/propenum.h"
#include "game/drawing.h"

/* @q Ufo(uid:Int):Obj (Function, Context)
   Access Ufo properties.
   Use as
   | ForEach Ufo Do ...
   or
   | With Ufo(n) Do ...
   @see int:index:group:ufoproperty|Ufo Properties, int:index:group:ufocommand|Ufo Commands */
IntValue*
IFUfoGet(IntArgBlock& args)
{
    int32_t id;
    args.checkArgumentCount(1, 1);
    if (!checkIntArg(id, args.getNext(), 0, 32767))
        return 0;

    if (!haveDisplayedTurn())
        return 0;

    GUfoType& con = getDisplayedTurn().getCurrentUniverse().ty_ufos;
    GUfoType::iterator it = con.getFirstAt(id);
    if (it != con.end() && it->getId() == id) {
        return new IntUfoContext(con, it);
    } else {
        return 0;
    }
}

IntContext*
IFUfoMake()
{
    if (!haveDisplayedTurn())
        return 0;

    GUfoType& con = getDisplayedTurn().getCurrentUniverse().ty_ufos;
    GUfoType::iterator it = con.begin();
    if (it != con.end()) {
        return new IntUfoContext(con, it);
    } else {
        return 0;
    }
}

int32_t
IFUfoDim(int32_t dim)
{
    return dim == 0 ? 1 : 32768;
}

/***************************** IntUfoContext *****************************/

namespace { enum { UfoProperty, UfoMethod }; }

static const TIntMapping ufo_mapping[] = {
    { "COLOR",          iupColorPCC,      UfoProperty, thInt },
    { "COLOR.EGA",      iupColorEGA,      UfoProperty, thInt },
    { "HEADING",        iupHeadingName,   UfoProperty, thString },
    { "HEADING$",       iupHeadingInt,    UfoProperty, thInt },
    { "ID",             iupId,            UfoProperty, thInt },
    { "ID2",            iupId2,           UfoProperty, thInt },
    { "INFO1",          iupInfo1,         UfoProperty, thString },
    { "INFO2",          iupInfo2,         UfoProperty, thString },
    { "KEEP",           iupKeepFlag,      UfoProperty, thBool },
    { "LASTSCAN",       iupLastScan,      UfoProperty, thInt },
    { "LOC.X",          iupLocX,          UfoProperty, thInt },
    { "LOC.Y",          iupLocY,          UfoProperty, thInt },
    { "MARK",           0,                UfoMethod,   thProcedure },
    { "MARKED",         iupMarked,        UfoProperty, thBool },
    { "MOVE.DX",        iupMoveDX,        UfoProperty, thInt },
    { "MOVE.DY",        iupMoveDY,        UfoProperty, thInt },
    { "NAME",           iupName,          UfoProperty, thString },
    { "RADIUS",         iupRadius,        UfoProperty, thInt },
    { "SPEED",          iupSpeedName,     UfoProperty, thString },
    { "SPEED$",         iupSpeedInt,      UfoProperty, thInt },
    { "TYPE",           iupType,          UfoProperty, thInt },
    { "UNMARK",         1,                UfoMethod,   thProcedure },
    { "VISIBLE.PLANET", iupVisiblePlanet, UfoProperty, thInt },
    { "VISIBLE.SHIP",   iupVisibleShip,   UfoProperty, thInt },
};

static const IntObjectProcedureValue::call_t ufo_methods[] = {
    IFObjMark,                  // 0
    IFObjUnmark,                // 1
};



IntUfoContext::IntUfoContext(GUfoType& con, GUfoType::iterator it)
    : con(con), it(it)
{ }

IntUfoContext::~IntUfoContext()
{ }

int32_t
IntUfoContext::lookup(const IntNameQuery& name)
{
    return lookupName(name, ufo_mapping, countof(ufo_mapping));
}

void
IntUfoContext::set(int32_t index, IntValue* value)
{
    switch (ufo_mapping[index].domain) {
     case UfoProperty:
        setUfoProperty(*it, IntUfoProperty(ufo_mapping[index].index), value);
        break;
     default:
        throw IntError::notAssignable();
    }
}

IntValue*
IntUfoContext::get(int32_t index)
{
    switch (ufo_mapping[index].domain) {
     case UfoProperty:
        return getUfoProperty(*it, IntUfoProperty(ufo_mapping[index].index));
     case UfoMethod:
        return new IntObjectProcedureValue(*it, getDisplayedTurn().getCurrentUniverse(), ufo_methods[ufo_mapping[index].index]);
     default:
        return 0;
    }
}

bool
IntUfoContext::next()
{
    GUfoType::iterator it2 = it;
    ++it2;
    if (it2 != con.end()) {
        it = it2;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntUfoContext::clone()
{
    return new IntUfoContext(con, it);
}

GObject*
IntUfoContext::getObject()
{
    return &*it;
}

string_t
IntUfoContext::toString(bool /*readable*/)
{
    return "#<ufo>";
}

void
IntUfoContext::store(IntTagNode& /*sv*/, Stream& /*aux*/)
{
    throw IntError::notSerializable();
}

void
IntUfoContext::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.enumMapping(ufo_mapping, countof(ufo_mapping));
}

/** Get Ufo property.
    \param ufo Ufo to query
    \param iup Property to query */
IntValue*
getUfoProperty(const GUfo& ufo, IntUfoProperty iup)
{
    switch (iup) {
     case iupColorEGA:
        /* @q Color.EGA:Int (Ufo Property)
           Ufo color code.
           This is the value reported by the host, as a value from the standard MS-DOS EGA palette.
           @see Color (Ufo Property) */
        return makeIntValue(ufo.getColorCode());
     case iupColorPCC:
        /* @q Color:Int (Ufo Property)
           Ufo color.
           This color is compatible to the {NewLine}, {NewMarker} etc. commands.
           @see Color (Ufo Property) */
        return makeIntValue(GDrawing::getUserColorFromColor(ufo.getColor()));
     case iupHeadingInt:
        /* @q Heading$:Int (Ufo Property)
           Heading, in degrees. EMPTY if Ufo does not move or heading is not known. */
        if (ufo.getHeading() >= 0)
            return makeIntValue(ufo.getHeading());
        else
            return 0;
     case iupHeadingName:
        /* @q Heading:Str (Ufo Property)
           Heading, as compass direction. */
        if (ufo.getHeading() >= 0)
            return makeStringValue(getHeadingName(ufo.getHeading()));
        else
            return 0;
     case iupId:
        /* @q Id:Int (Ufo Property)
           Ufo Id. */
        return makeIntValue(ufo.getId());
     case iupId2:
        /* @q Id2:Int (Ufo Property)
           Real Id number, or 0.
           Some add-ons send their own objects, which may have different Id numbers,
           to Ufos to make them visible to players.
           This field is to support these add-ons.
           Currently, this field is used with PHost's wormholes,
           where it holds the real Id number of the wormhole,
           using the usual rules (even Id = entry, odd Id = exit).
           @assignable */
        return makeIntValue(ufo.getRealId());
     case iupInfo1:
        /* @q Info1:Str (Ufo Property), Info2:Str (Ufo Property)
           Description of this Ufo. */
        return makeStringValue(ufo.getInfo1());
     case iupInfo2:
        return makeStringValue(ufo.getInfo2());
     case iupKeepFlag:
        /* @q Keep:Bool (Ufo Property)
           True to keep this Ufo in the database.
           Defaults to False, i.e. the Ufo is only displayed when sent by the host.
           @assignable */
        return makeBoolValue(ufo.isStoredInHistory());
     case iupLastScan:
        /* @q LastScan:Int (Ufo Property)
           Turn when Ufo was last scanned. */
        return makeIntValue(ufo.getLastTurn());
     case iupLocX:
        /* @q Loc.X:Int (Ufo Property)
           X location of Ufo center. */
        return makeIntValue(ufo.getPos().x);
     case iupLocY:
        /* @q Loc.Y:Int (Ufo Property)
           Y location of Ufo center. */
        return makeIntValue(ufo.getPos().y);
     case iupMarked:
        /* @q Marked:Bool (Ufo Property)
           True if Ufo is marked. */
        return makeBoolValue(ufo.isMarked());
     case iupMoveDX:
        /* @q Move.DX:Int (Ufo Property)
           Estimated/average per-turn movement in X direction.
           Used to update guessed positions when the Ufo is not scanned.
           @assignable */
        return makeIntValue(ufo.getMovementVector().x);
     case iupMoveDY:
        /* @q Move.DY:Int (Ufo Property)
           Estimated/average per-turn movement in Y direction.
           Used to update guessed positions when the Ufo is not scanned.
           @assignable */
        return makeIntValue(ufo.getMovementVector().y);
     case iupName:
        /* @q Name:Str (Ufo Property)
           Name of Ufo. */
        return makeStringValue(ufo.getName(GObject::PlainName));
     case iupRadius:
        /* @q Radius:Int (Ufo Property)
           Radius of Ufo in ly. */
        return makeIntValue(ufo.getRadius());
     case iupSpeedInt:
        /* @q Speed$:Int (Ufo Property)
           Speed (warp factor). */
        if (ufo.getSpeed() >= 0)
            return makeIntValue(ufo.getSpeed());
        else
            return 0;
     case iupSpeedName:
        /* @q Speed:Str (Ufo Property)
           Speed, as human-readable string. */
        if (ufo.getSpeed() >= 0)
            return makeStringValue(format("Warp %d", ufo.getSpeed()));
        else
            return 0;
     case iupType:
        /* @q Type:Int (Ufo Property)
           Type of Ufo.
           This is an integer reported by the add-on providing the Ufo,
           identifying the Ufo type. */
        return makeIntValue(ufo.getTypeCode());
     case iupVisiblePlanet:
        /* @q Visible.Planet:Int (Ufo Property)
           Distance from which Ufo can be seen from a planet, in ly. */
        return makeIntValue(ufo.getPlanetRange());
     case iupVisibleShip:
        /* @q Visible.Ship:Int (Ufo Property)
           Distance from which Ufo can be seen from a ship, in ly. */
        return makeIntValue(ufo.getShipRange());
    }
    return 0;
}

/** Set Ufo property.
    \param ufo Ufo to modify
    \param iup Property to modify
    \param value Value to set */
void
setUfoProperty(GUfo& ufo, const IntUfoProperty iup, IntValue*const value)
{
    int32_t iv;
    switch (iup) {
     case iupId2:
        if (checkIntArg(iv, value)) {
            ufo.setRealId(iv);
        }
        break;

     case iupKeepFlag:
        iv = getBoolValue(value);
        if (iv >= 0) {
            ufo.setIsStoredInHistory(iv != 0);
        }
        break;

     case iupMoveDX:
     case iupMoveDY:
        if (checkIntArg(iv, value, -500, 500)) {
            GPoint pt = ufo.getMovementVector();
            if (iup == iupMoveDX) {
                pt.x = iv;
            } else {
                pt.y = iv;
            }
            ufo.setMovementVector(pt);
        }
        break;

     default:
        throw IntError::notAssignable();
    }
}
