/**
  *  \file u/t_game_coord.cc
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/coord.h"
#include "u/t_game.h"

void
GameCoordTestSuite::testBasics()
{
    // Essentially, we can only test construction and comparison here,
    // as everything else needs a chart geometry.
    GPoint a(10, 20);
    GPoint b(20, 30);
    GPoint c(10, 20);

    TS_ASSERT(a != b);
    TS_ASSERT(a == c);
    TS_ASSERT(b != a);
    TS_ASSERT(b != c);
    TS_ASSERT(c == a);
    TS_ASSERT(c != b);

    TS_ASSERT_EQUALS(a.x, 10);
    TS_ASSERT_EQUALS(a.y, 20);
    TS_ASSERT_EQUALS(b.x, 20);
    TS_ASSERT_EQUALS(b.y, 30);
    TS_ASSERT_EQUALS(c.x, 10);
    TS_ASSERT_EQUALS(c.y, 20);
}

void
GameCoordTestSuite::testFlat()
{
    // Configure map to not-wrapped, standard size
    chart_conf.setConfig(GChartConfiguration::Flat, GPoint(2000, 2000), GPoint(2000, 2000));
    TS_ASSERT(chart_conf.getMinimumCoordinates() == GPoint(1000, 1000));
    TS_ASSERT(chart_conf.getMaximumCoordinates() == GPoint(3000, 3000));

    // Normalizing does not modify points
    TS_ASSERT(GPoint(10, 20).getSimpleCanonicalLocation() == GPoint(10, 20));
    TS_ASSERT(GPoint(10, 20).getCanonicalLocation() == GPoint(10, 20));
    TS_ASSERT(GPoint(10, 20).getSimpleNearestAlias(GPoint(2900, 2900)) == GPoint(10, 20));
    TS_ASSERT(GPoint(10, 20).getSimpleNearestAlias(GPoint(1100, 2900)) == GPoint(10, 20));
    TS_ASSERT(GPoint(10, 20).getSimpleNearestAlias(GPoint(2900, 1100)) == GPoint(10, 20));
    TS_ASSERT(GPoint(10, 20).getSimpleNearestAlias(GPoint(1100, 1100)) == GPoint(10, 20));

    TS_ASSERT(GPoint(10, 20).isOnMap());

    // Sector numbers known for all points in [1000,3000)
    // Sectors are numbered
    //   100 110 120 130 ... 190 200 210 ...
    //   101 111 121 131 ... 191 201 211 ...
    //   ... ... ...
    //   109 119 129 139 ... 199 209 219 ...
    //   300 310 320 330 ... 390 400 410 ...
    //   ... ... ...
    //   309 319 329 339 ... 399 409 419 ...
    TS_ASSERT_EQUALS(GPoint(1000, 1000).getSectorNumber(), 309);
    TS_ASSERT_EQUALS(GPoint(1099, 1099).getSectorNumber(), 309);
    TS_ASSERT_EQUALS(GPoint(1111, 1000).getSectorNumber(), 319);
    TS_ASSERT_EQUALS(GPoint(1222, 1000).getSectorNumber(), 329);
    TS_ASSERT_EQUALS(GPoint(1111, 1111).getSectorNumber(), 318);

    TS_ASSERT_EQUALS(GPoint(2000, 1000).getSectorNumber(), 409);
    TS_ASSERT_EQUALS(GPoint(1000, 2000).getSectorNumber(), 109);
    TS_ASSERT_EQUALS(GPoint(2000, 2000).getSectorNumber(), 209);

    TS_ASSERT_EQUALS(GPoint(2999, 2999).getSectorNumber(), 290);

    TS_ASSERT_EQUALS(GPoint(1500, 1000).getSectorNumber(), 359);
    TS_ASSERT_EQUALS(GPoint(1500, 1100).getSectorNumber(), 358);
    TS_ASSERT_EQUALS(GPoint(1500, 1200).getSectorNumber(), 357);
    TS_ASSERT_EQUALS(GPoint(1500, 1300).getSectorNumber(), 356);
    TS_ASSERT_EQUALS(GPoint(1500, 1400).getSectorNumber(), 355);
    TS_ASSERT_EQUALS(GPoint(1500, 1500).getSectorNumber(), 354);
    TS_ASSERT_EQUALS(GPoint(1500, 1600).getSectorNumber(), 353);
    TS_ASSERT_EQUALS(GPoint(1500, 1700).getSectorNumber(), 352);
    TS_ASSERT_EQUALS(GPoint(1500, 1800).getSectorNumber(), 351);

    // Check that parsed sector numbers are centered within their respective sector
    for (int i = 100; i < 500; ++i) {
        GPoint p;
        TS_ASSERT(p.parseSectorNumber(i));
        TS_ASSERT_EQUALS(p.getSectorNumber(), i);
        TS_ASSERT_EQUALS(p.x % 100, 50);
        TS_ASSERT_EQUALS(p.y % 100, 50);
    }
    TS_ASSERT(!GPoint().parseSectorNumber(0));
    TS_ASSERT(!GPoint().parseSectorNumber(-1));
    TS_ASSERT(!GPoint().parseSectorNumber(99));
    TS_ASSERT(!GPoint().parseSectorNumber(500));
    TS_ASSERT(!GPoint().parseSectorNumber(501));

    // Some out-of-range values
    TS_ASSERT_EQUALS(GPoint(999,  999).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1999, 999).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(999,  1999).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1999, 3001).getSectorNumber(), 0);
}

void
GameCoordTestSuite::testFlatSmall()
{
    // Configure map to not-wrapped, smaller size
    chart_conf.setConfig(GChartConfiguration::Flat, GPoint(2000, 2000), GPoint(1000, 1000));
    TS_ASSERT(chart_conf.getMinimumCoordinates() == GPoint(1500, 1500));
    TS_ASSERT(chart_conf.getMaximumCoordinates() == GPoint(2500, 2500));

    // Normalizing still does not modify points
    TS_ASSERT(GPoint(10, 20).getSimpleCanonicalLocation() == GPoint(10, 20));
    TS_ASSERT(GPoint(10, 20).getCanonicalLocation() == GPoint(10, 20));
    TS_ASSERT(GPoint(10, 20).getSimpleNearestAlias(GPoint(2900, 2900)) == GPoint(10, 20));
    TS_ASSERT(GPoint(10, 20).getSimpleNearestAlias(GPoint(1100, 2900)) == GPoint(10, 20));
    TS_ASSERT(GPoint(10, 20).getSimpleNearestAlias(GPoint(2900, 1100)) == GPoint(10, 20));
    TS_ASSERT(GPoint(10, 20).getSimpleNearestAlias(GPoint(1100, 1100)) == GPoint(10, 20));

    TS_ASSERT(GPoint(10, 20).isOnMap());

    // Sector numbers still known for all points on map:
    TS_ASSERT_EQUALS(GPoint(1000, 1000).getSectorNumber(), 309);
    TS_ASSERT_EQUALS(GPoint(1099, 1099).getSectorNumber(), 309);
    TS_ASSERT_EQUALS(GPoint(1111, 1000).getSectorNumber(), 319);
    TS_ASSERT_EQUALS(GPoint(1222, 1000).getSectorNumber(), 329);
    TS_ASSERT_EQUALS(GPoint(1111, 1111).getSectorNumber(), 318);

    TS_ASSERT_EQUALS(GPoint(2000, 1000).getSectorNumber(), 409);
    TS_ASSERT_EQUALS(GPoint(1000, 2000).getSectorNumber(), 109);
    TS_ASSERT_EQUALS(GPoint(2000, 2000).getSectorNumber(), 209);

    TS_ASSERT_EQUALS(GPoint(2999, 2999).getSectorNumber(), 290);

    TS_ASSERT_EQUALS(GPoint(1500, 1000).getSectorNumber(), 359);
    TS_ASSERT_EQUALS(GPoint(1500, 1100).getSectorNumber(), 358);
    TS_ASSERT_EQUALS(GPoint(1500, 1200).getSectorNumber(), 357);
    TS_ASSERT_EQUALS(GPoint(1500, 1300).getSectorNumber(), 356);
    TS_ASSERT_EQUALS(GPoint(1500, 1400).getSectorNumber(), 355);
    TS_ASSERT_EQUALS(GPoint(1500, 1500).getSectorNumber(), 354);
    TS_ASSERT_EQUALS(GPoint(1500, 1600).getSectorNumber(), 353);
    TS_ASSERT_EQUALS(GPoint(1500, 1700).getSectorNumber(), 352);
    TS_ASSERT_EQUALS(GPoint(1500, 1800).getSectorNumber(), 351);

    // Check that parsed sector numbers are centered within their respective sector
    for (int i = 100; i < 500; ++i) {
        GPoint p;
        TS_ASSERT(p.parseSectorNumber(i));
        TS_ASSERT_EQUALS(p.getSectorNumber(), i);
        TS_ASSERT_EQUALS(p.x % 100, 50);
        TS_ASSERT_EQUALS(p.y % 100, 50);
    }
    TS_ASSERT(!GPoint().parseSectorNumber(0));
    TS_ASSERT(!GPoint().parseSectorNumber(-1));
    TS_ASSERT(!GPoint().parseSectorNumber(99));
    TS_ASSERT(!GPoint().parseSectorNumber(500));
    TS_ASSERT(!GPoint().parseSectorNumber(501));

    // Some out-of-range values
    TS_ASSERT_EQUALS(GPoint(999,  999).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1999, 999).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(999,  1999).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1999, 3001).getSectorNumber(), 0);
}

void
GameCoordTestSuite::testFlatOffset()
{
    // Configure map to not-wrapped, smaller size and not centered at 2000,2000
    chart_conf.setConfig(GChartConfiguration::Flat, GPoint(1750, 2500), GPoint(1500, 1000));
    TS_ASSERT_EQUALS(chart_conf.getMinimumCoordinates().x, 1000);
    TS_ASSERT_EQUALS(chart_conf.getMinimumCoordinates().y, 2000);
    TS_ASSERT_EQUALS(chart_conf.getMaximumCoordinates().x, 2500);
    TS_ASSERT_EQUALS(chart_conf.getMaximumCoordinates().y, 3000);

    // Normalizing still does not modify points
    TS_ASSERT(GPoint(10, 20).getSimpleCanonicalLocation() == GPoint(10, 20));
    TS_ASSERT(GPoint(10, 20).getCanonicalLocation() == GPoint(10, 20));
    TS_ASSERT(GPoint(10, 20).getSimpleNearestAlias(GPoint(2900, 2900)) == GPoint(10, 20));
    TS_ASSERT(GPoint(10, 20).getSimpleNearestAlias(GPoint(1100, 2900)) == GPoint(10, 20));
    TS_ASSERT(GPoint(10, 20).getSimpleNearestAlias(GPoint(2900, 1100)) == GPoint(10, 20));
    TS_ASSERT(GPoint(10, 20).getSimpleNearestAlias(GPoint(1100, 1100)) == GPoint(10, 20));

    TS_ASSERT(GPoint(10, 20).isOnMap());

    // Sector numbers still known for all points on map, but offset
    // relative to new center at 1750,2500 instead of 2000,2000.
    // Those are now out-of-range:
    TS_ASSERT_EQUALS(GPoint(1000, 1000).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1099, 1099).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1111, 1000).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1222, 1000).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1111, 1111).getSectorNumber(), 0);

    TS_ASSERT_EQUALS(GPoint(2000, 1000).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1000, 2000).getSectorNumber(), 324);
    TS_ASSERT_EQUALS(GPoint(2000, 2000).getSectorNumber(), 424);

    TS_ASSERT_EQUALS(GPoint(2999, 2999).getSectorNumber(), 0);

    TS_ASSERT_EQUALS(GPoint(1500, 1000).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1500, 1100).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1500, 1200).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1500, 1300).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1500, 1400).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1500, 1500).getSectorNumber(), 379);
    TS_ASSERT_EQUALS(GPoint(1500, 1600).getSectorNumber(), 378);
    TS_ASSERT_EQUALS(GPoint(1500, 1700).getSectorNumber(), 377);
    TS_ASSERT_EQUALS(GPoint(1500, 1800).getSectorNumber(), 376);

    TS_ASSERT_EQUALS(GPoint(999,   999).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1999,  999).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(999,  1999).getSectorNumber(), 325);
    TS_ASSERT_EQUALS(GPoint(1999, 3001).getSectorNumber(), 224);

    // Check that parsed sector numbers are centered within their respective sector
    // Sectors are now centered around (xx00,xx50), not (xx00,xx50)
    for (int i = 100; i < 500; ++i) {
        GPoint p;
        TS_ASSERT(p.parseSectorNumber(i));
        TS_ASSERT_EQUALS(p.getSectorNumber(), i);
        TS_ASSERT_EQUALS(p.x % 100, 0);
        TS_ASSERT_EQUALS(p.y % 100, 50);
    }
    TS_ASSERT(!GPoint().parseSectorNumber(0));
    TS_ASSERT(!GPoint().parseSectorNumber(-1));
    TS_ASSERT(!GPoint().parseSectorNumber(99));
    TS_ASSERT(!GPoint().parseSectorNumber(500));
    TS_ASSERT(!GPoint().parseSectorNumber(501));
}

void
GameCoordTestSuite::testWrapped()
{
    // Configure map to wrapped, standard size
    chart_conf.setConfig(GChartConfiguration::Wrapped, GPoint(2000, 2000), GPoint(2000, 2000));

    // Normalizing
    TS_ASSERT(GPoint(10, 20).getSimpleCanonicalLocation() == GPoint(2010, 2020));
    TS_ASSERT(GPoint(10, 20).getCanonicalLocation() == GPoint(2010, 2020));
    TS_ASSERT(GPoint(10, 20).getSimpleNearestAlias(GPoint(2900, 2900)) == GPoint(2010, 2020));
    TS_ASSERT(GPoint(10, 20).getSimpleNearestAlias(GPoint(1100, 2900)) == GPoint(2010, 2020));
    TS_ASSERT(GPoint(10, 20).getSimpleNearestAlias(GPoint(2900, 1100)) == GPoint(2010, 2020));
    TS_ASSERT(GPoint(10, 20).getSimpleNearestAlias(GPoint(1100, 1100)) == GPoint(2010, 2020));

    TS_ASSERT(GPoint(1010, 1020).getSimpleNearestAlias(GPoint(2900, 2900)) == GPoint(3010, 3020));
    TS_ASSERT(GPoint(1010, 1020).getSimpleNearestAlias(GPoint(1100, 2900)) == GPoint(1010, 3020));
    TS_ASSERT(GPoint(1010, 1020).getSimpleNearestAlias(GPoint(2900, 1100)) == GPoint(3010, 1020));
    TS_ASSERT(GPoint(1010, 1020).getSimpleNearestAlias(GPoint(1100, 1100)) == GPoint(1010, 1020));

    TS_ASSERT(!GPoint(10, 20).isOnMap());

    // Sector numbers known for all points in [1000,3000), same as in testFlat()
    TS_ASSERT_EQUALS(GPoint(1000, 1000).getSectorNumber(), 309);
    TS_ASSERT_EQUALS(GPoint(1099, 1099).getSectorNumber(), 309);
    TS_ASSERT_EQUALS(GPoint(1111, 1000).getSectorNumber(), 319);
    TS_ASSERT_EQUALS(GPoint(1222, 1000).getSectorNumber(), 329);
    TS_ASSERT_EQUALS(GPoint(1111, 1111).getSectorNumber(), 318);

    TS_ASSERT_EQUALS(GPoint(2000, 1000).getSectorNumber(), 409);
    TS_ASSERT_EQUALS(GPoint(1000, 2000).getSectorNumber(), 109);
    TS_ASSERT_EQUALS(GPoint(2000, 2000).getSectorNumber(), 209);

    TS_ASSERT_EQUALS(GPoint(2999, 2999).getSectorNumber(), 290);

    TS_ASSERT_EQUALS(GPoint(1500, 1000).getSectorNumber(), 359);
    TS_ASSERT_EQUALS(GPoint(1500, 1100).getSectorNumber(), 358);
    TS_ASSERT_EQUALS(GPoint(1500, 1200).getSectorNumber(), 357);
    TS_ASSERT_EQUALS(GPoint(1500, 1300).getSectorNumber(), 356);
    TS_ASSERT_EQUALS(GPoint(1500, 1400).getSectorNumber(), 355);
    TS_ASSERT_EQUALS(GPoint(1500, 1500).getSectorNumber(), 354);
    TS_ASSERT_EQUALS(GPoint(1500, 1600).getSectorNumber(), 353);
    TS_ASSERT_EQUALS(GPoint(1500, 1700).getSectorNumber(), 352);
    TS_ASSERT_EQUALS(GPoint(1500, 1800).getSectorNumber(), 351);

    // Check that parsed sector numbers are centered within their respective sector
    for (int i = 100; i < 500; ++i) {
        GPoint p;
        TS_ASSERT(p.parseSectorNumber(i));
        TS_ASSERT_EQUALS(p.getSectorNumber(), i);
        TS_ASSERT_EQUALS(p.x % 100, 50);
        TS_ASSERT_EQUALS(p.y % 100, 50);
        TS_ASSERT_EQUALS(p, p.getSimpleCanonicalLocation());
        TS_ASSERT_EQUALS(p, p.getCanonicalLocation());
    }
    TS_ASSERT(!GPoint().parseSectorNumber(0));
    TS_ASSERT(!GPoint().parseSectorNumber(-1));
    TS_ASSERT(!GPoint().parseSectorNumber(99));
    TS_ASSERT(!GPoint().parseSectorNumber(500));
    TS_ASSERT(!GPoint().parseSectorNumber(501));

    // Some out-of-range values
    TS_ASSERT_EQUALS(GPoint(999,  999).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1999, 999).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(999,  1999).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1999, 3001).getSectorNumber(), 0);
}

void
GameCoordTestSuite::testWrappedSmall()
{
    // Configure map to wrapped, small size
    chart_conf.setConfig(GChartConfiguration::Wrapped, GPoint(2000, 2000), GPoint(1000, 1000));

    // Sector numbers known for all points in [1500,2500), but still numbered as in normal map.
    // Parts are out of range.
    TS_ASSERT_EQUALS(GPoint(1000, 1000).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1099, 1099).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1111, 1000).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1222, 1000).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1111, 1111).getSectorNumber(), 0);

    TS_ASSERT_EQUALS(GPoint(2000, 1000).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1000, 2000).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(2000, 2000).getSectorNumber(), 209);

    TS_ASSERT_EQUALS(GPoint(2999, 2999).getSectorNumber(), 0);

    TS_ASSERT_EQUALS(GPoint(1500, 1000).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1500, 1100).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1500, 1200).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1500, 1300).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1500, 1400).getSectorNumber(), 0);
    TS_ASSERT_EQUALS(GPoint(1500, 1500).getSectorNumber(), 354);
    TS_ASSERT_EQUALS(GPoint(1500, 1600).getSectorNumber(), 353);
    TS_ASSERT_EQUALS(GPoint(1500, 1700).getSectorNumber(), 352);
    TS_ASSERT_EQUALS(GPoint(1500, 1800).getSectorNumber(), 351);
}
